#include <iostream>
#include <interval.hpp>
#include "itaylor.hpp"   // interval taylor arithmetic 

using namespace std;
using namespace cxsc;
using namespace taylor;

typedef itaylor (*itayl_FctPtr)(const itaylor&);	// function pointer

// ---------------------------------------------------------------------

interval integral (itayl_FctPtr f, const real &a, const real &b, const real &abs_err, bool error)
{
   int      p = 5;        // Order of expansion for taylor arithmetic
   interval z(a,b);       // Interval to include the point of expansion
   int      n = 2,j;
   interval h;            // length of each subinterval
   itaylor  x(p,z);       // function variables
   interval sum(0.0);     // Summation of the 4th derivatives (error bound)
 	
   // initial step: error bound for ordinary Simpson rule 	
   h = (interval(b)-interval(a))/n;   
   interval R     = get_j_derivative(f(x),4)*4.0*h*sqr(sqr(h))/15.0;
   interval R_old = R;

   while( ( diam(R) > abs_err ) && (diam(R) <= diam(R_old) ) )
   // iterative refinement of the grid
   // compute error bound for composite Simpson rule
   {
      n *= 2;
      h = (interval(b)-interval(a))/n;
      real x2i_minus_2, x2i;  // for f[x_2i-2,x_2i]

      R_old = R;

      for (int i=1;i<=n/2;i++)
      {
         j = 2*(i-1);
         itaylor x( p , interval(a) +  h*interval(j,j+2));	
         sum += get_j_derivative(f(x),4);
      }
	
      R = 4.0*h*sqr(sqr(h))*sum/15.0;	// computation of the remainder term
   }
	
   // ----------------------------------------------------------------------
   
   // compute sum of function values 
   interval J;
   interval xk;		     // means x[k]
   real si;                  // weight of the Simpsons sum
   idotprecision accu(0.0);  // accurate dot product

   for (int k=0;k<=n;k++)
   {
      xk = a+k*h;
      if ( (k==0) || (k==n) )
         si = 1.0;
      else if ( k%2 == 0 )
         si = 2.0;
      else 
         si = 4.0;
      // accurate summation 
      accumulate(accu, si, get_j_derivative(f(var_itaylor(5,xk)),0) );
    }
	
    J = h * rnd(accu) / 3.0;

   // ----------------------------------------------------------------------

   interval res = R + J;

   // ---------------------------------------------------------------------

   // test if result meets required accuracy
   if ( diam(res) <= abs_err ) 
      error = 0;
   else
      error = 1;
	
   return res;
}

// ---------------------------------------------------------------------

itaylor f ( const itaylor &X )   // sample function
{ 
   return 2 * X * exp(sqr(X)) * sin(exp(sqr(X))) ; 
}

// ---------------------------------------------------------------------

int main (void)
{
   real a = 0, b = 2.0;         // a, b integration bounds
   real abs_err = 1e-10;        // upper bound for absolute error
   bool error;                  // error indicator
	
   interval res;
   res = integral( f, a, b, abs_err, error);
   cout << SetPrecision(17,10) << Scientific << endl;
   cout << "Integration of f(x) = 2 * X * exp(sqr(X)) * sin( exp(sqr(X)))" << endl;
   cout << "Simpson Rule: " << res << endl;
   if (error == 1) cout << "Warning: Insufficient accuracy." << endl;
   return 0;
}

/* OUTPUT:

Integration of f(x) = 2 * X * exp(sqr(X)) * sin( exp(sqr(X)))
Simpson Rule: [9.1096403927E-001,9.1096403929E-001]
*/
